"""
Represents a single ``Column`` of a :class:`~snowmobile.core.snowframe.SnowFrame`,
primarily for the purpose of carrying out/tracking operations performed
on a given field in order to standardize or re-format it prior to loading
into a table.
"""
from __future__ import annotations

import itertools
import re
import string
from contextlib import contextmanager
from typing import Optional

from . import Generic


class Column(Generic):
    """A single column within a :class:`~snowmobile.core.snowframe.SnowFrame`.

    Attributes:
        original (str):
            Original column name.
        current (str):
            Current version of the column name.
        prior (str):
            Prior version of the column name (version `n-1`).
        src (str):
            Column source (original ``df`` or added by ``snowmobile``).

    """

    _EXCLUDE_CHARS = list(
        itertools.chain.from_iterable(
            (
                [c for c in string.punctuation if c != "_"],
                [w for w in string.whitespace],
            )
        )
    )

    def __init__(
        self,
        original: str,
        current: Optional[str] = None,
        prior: Optional[str] = None,
        src: Optional[str] = None,
    ):
        super().__init__()
        
        self.original = original
        """str: Original column name."""
        
        self.src = src or "original"
        """str: Source of column; 'df' if from source DataFrame, 'snowmobile' otherwise."""
        
        self.current = current or self.original
        """str: Current column name."""
        
        self.prior = prior or self.current
        """str: Prior (version of) column name."""

    @contextmanager
    def update(self):
        """Migrate from prior to current context within this context."""
        try:
            self.prior = self.current
            yield self
        finally:
            return self.current

    def lower(self) -> str:
        """Lower case column."""
        with self.update() as s:
            s.current = s.current.lower()
        return self.current

    def upper(self) -> str:
        """Upper case column."""
        with self.update() as s:
            s.current = s.current.upper()
        return self.current

    @staticmethod
    def dedupe(current: str, char: Optional[str] = None) -> str:
        """Dedupes consecutive characters within a string.

        Note:
            *   Must iterate through matches and perform replacements in the
                order of the **largest to the smallest by number of characters**;
                this is to avoid altering the matches found before replacing them.

        Args:
            current (str):
                String containing characters to dedupe.
            char (str):
                Character to dedupe.

        """
        matches = re.findall(f"{char}+", current)
        for match in reversed(matches):
            current = current.replace(match, char)
        return current

    def reformat(
        self, fill_char: Optional[str] = None, dedupe_special: bool = True
    ) -> str:
        """Reformat column for a load to the database.

        Args:
            fill_char (str):
                Character to replace special characters and whitespace with;
                defaults to `_`.
            dedupe_special (bool):
                Dedupe consecutive special characters; defaults to `True`.

        """
        with self.update() as s:
            fill_char = fill_char or "_"
            to_swap = {k: fill_char for k in self._EXCLUDE_CHARS}
            to_swap['_'] = '_'
            for char, swap_char in to_swap.items():
                s.current = s.current.replace(char, swap_char)
            if dedupe_special:
                s.current = s.dedupe(current=s.current, char=fill_char)
        return self.current

    def __eq__(self, other: Column) -> bool:
        return all(
            (
                self.original == other.original,
                self.src == other.src,
                self.current == other.current,
                self.prior == other.prior,
            )
        )

    def __str__(self) -> str:
        return f"Column(current='{self.current}')"

    def __repr__(self) -> str:
        return f"Column(original='{self.original}', current='{self.current}')"



"""
snowmobile exception classes.
"""
import time
from typing import Dict, List, Optional

from pandas.io.sql import DatabaseError as pdDataBaseError
from snowflake.connector.errors import DatabaseError as sfDatabaseError
from snowflake.connector.errors import ProgrammingError


class Error(Exception):
    """Base Generic exception class.

    Args:
        msg (Optional[str]):
            Error message.
        errno (Optional[int]):
            Error number.
        nm (Optional[str]):
            Globally unique name for an exception being raised; used for
            InternalExceptions and statement names when QA failures occur.
        to_raise (Optional[bool]):
            Indicates that the exception should be raised before exiting the
            current context.

    Attributes:
        tmstmp (int):
            Unix instantiation timestamp of the exception (in seconds).
        raised (bool):
            Indicator of whether or not the instance has already been raised;
            used for exception chaining from Statement -> Script.

    """

    def __init__(
        self,
        msg: Optional[str] = None,
        errno: Optional[int] = None,
        nm: Optional[str] = None,
        to_raise: Optional[bool] = False,
    ):
        self.tmstmp: int = int(time.time())
        self.msg = (msg or str()).strip("\n")
        self.errno = errno
        self.nm = nm
        self.to_raise: bool = to_raise
        self.raised: bool = False

    def __str__(self):
        """Default error message."""
        return f"{self.msg}" if self.msg else f"Exception encountered"

    def escalate(self):
        """Raises error."""
        self.raised = True
        raise self

    @staticmethod
    def format_error_args(
        prefix: Optional[str] = None,
        sep: Optional[str] = None,
        lines: Optional[int] = None,
        _filter: bool = True,
        **kwargs: Dict[str, str],
    ) -> str:
        """Formats a dictionary of arguments into an aligned/indented error msg.

        Placed below primary msg such that a primary msg of 'This is a __ error.'
        combined with the returned value from this method pr with
        kwargs={'argument-description1': 'argument-value', 'arg2-desc': 'arg2-value'}
        would produce the following error message:
        
            ```
            This is a ClassName error.
                argument-description: argument-value
                           arg2-desc: arg2-value
            ```

        Args:
            prefix (str):
                Character to prefix bullets with; defaults to '\t'.
            sep (str):
                Character to separate arguments/values with; defaults to ':'.
            lines (int):
                Number of lines to include between arguments; defaults to 1.
            _filter (bool):
                Indicator of whether to filter out key/value pairs that contain
                empty values; defaults to `True`.
            **kwargs:
                Argument keys and values to be converted into a list.

        Returns (str):
            Formatted arguments as a string.

        """
        prefix = prefix or "\t"
        line_sep = "\n" * (lines or 1)
        sep = sep or ": "

        if _filter:
            kwargs = {k: v for k, v in kwargs.items() if v}

        longest = max(len(k) for k in kwargs)
        args = [
            f"{prefix}{k.rjust(len(k) + (longest-len(k)))}{sep}{v}"
            for k, v in kwargs.items()
        ]

        return line_sep.join(args)


class InternalError(Error):
    """Exceptions raised from an internal state that should not exist.

    Debugging/testing Exceptions only.

    """

    def __init__(
        self,
        msg: Optional[str] = None,
        errno: Optional[int] = None,
        nm: Optional[str] = None,
        to_raise: Optional[bool] = False,
    ):
        super().__init__(nm=nm, msg=msg, errno=errno, to_raise=to_raise)

    def __str__(self):
        """InternalError message."""
        str_args = self.format_error_args(
            _filter=True,
            **{"name": f"'{self.nm}'", "msg": self.msg, "errno": self.errno},
        ).strip("\n")
        return f"""
An internal exception was raised.
{str_args}
"""


class StatementInternalError(InternalError):
    """Statement-specific internal errors."""

    def __init__(
        self,
        msg: Optional[str] = None,
        errno: Optional[int] = None,
        nm: Optional[str] = None,
        to_raise: Optional[bool] = False,
    ):
        super().__init__(nm=nm, msg=msg, errno=errno, to_raise=to_raise)


class StatementPostProcessingError(Error):
    """Exceptions encountered in post-processing invoked by `s.process()`.

    Indicates a non-database error occurred in ``.process()`` method of a
    derived :class:`~snowmobile.core.statement.Statement` class.

    """

    def __init__(
        self,
        msg: Optional[str] = None,
        errno: Optional[int] = None,
        to_raise: Optional[bool] = False,
    ):
        super().__init__(msg=msg, errno=errno, to_raise=to_raise)

    def __str__(self):
        """StatementPostProcessingError message."""
        return (
            f"An error was encountered during post-processing '{self.msg}'"
            if self.msg
            else f"Post-processing error encountered"
        )


class QAFailure(Error):
    """Boolean eval of a statement's `outcome` attribute is *False*.

    Indicates the logic invoked by the ``.process()`` method
    completed successfully and has set the ``outcome`` attribute
    to False.

    Args:
        nm (str):
            Name name of QA statement.
        desc (str):
            Object-specific exception message to display.
        idx (int):
            Index of the statement that failed its QA check.

    """

    def __init__(
        self,
        nm: str,
        msg: str,
        idx: int,
        desc: Optional[str] = None,
        errno: Optional[int] = None,
        to_raise: Optional[bool] = False,
    ):
        super().__init__(msg=msg, errno=errno, nm=nm, to_raise=to_raise)
        self.desc = desc
        self.idx = idx

    def __str__(self):
        """QAFailure message."""
        str_args = self.format_error_args(
            _filter=True,
            **{
                "name": f"'{self.nm}' (s{self.idx})",
                "msg": self.msg,
                "user-description": self.desc,
            },
        ).strip("\n")
        return f"""
A configured QA check did not pass its validation:
{str_args}
"""


class QAEmptyFailure(QAFailure):
    """:class:`~snowmobile.core.qa.Empty` specific failure."""

    def __init__(
        self,
        nm: str,
        msg: str,
        idx: int,
        desc: Optional[str] = None,
        errno: Optional[int] = None,
        to_raise: Optional[bool] = False,
    ):
        super().__init__(
            nm=nm, msg=msg, idx=idx, desc=desc, errno=errno, to_raise=to_raise
        )


class QADiffFailure(QAFailure):
    """:class:`~snowmobile.core.qa.Diff` specific failure."""

    def __init__(
        self,
        nm: str,
        msg: str,
        idx: int,
        desc: Optional[str] = None,
        errno: Optional[int] = None,
        to_raise: Optional[bool] = False,
    ):
        super().__init__(
            nm=nm, msg=msg, idx=idx, desc=desc, errno=errno, to_raise=to_raise
        )


class SnowFrameInternalError(InternalError):
    """class:`~snowmobile.core.snowframe.SnowFrame` specific internal error."""

    def __init__(
        self,
        msg: Optional[str] = None,
        errno: Optional[int] = None,
        nm: Optional[str] = None,
        to_raise: Optional[bool] = False,
    ):
        super().__init__(nm=nm, msg=msg, errno=errno, to_raise=to_raise)

    def __str__(self):
        """SnowFrameInternalError message."""
        str_args = self.format_error_args(
            _filter=True, **{"name": self.nm, "msg": self.msg}
        ).strip("\n")
        return f"""
{str_args}
"""


class StatementNotFoundError(Error):
    """Invalid statement name or index has been pr."""

    def __init__(
        self,
        nm: str,
        statements: List[str] = None,
        msg: Optional[str] = None,
        errno: Optional[int] = None,
        to_raise: Optional[bool] = False,
    ):
        super().__init__(nm=nm, msg=msg, errno=errno, to_raise=to_raise)
        self.st = statements

    def __str__(self):
        """StatementNotFoundError message."""
        statements = ", ".join(f"'{s}'" for s in self.st) if self.st else ""
        str_args = self.format_error_args(
            _filter=True,
            **{
                "name-pr": f"'{self.nm}'",
                "msg": self.msg,
                "errno": self.errno,
                "names-found": statements,
            },
        ).strip("\n")
        return f"""
Statement name or index, `{self.nm}`, is not found in the script.
{str_args}
"""


class DuplicateTagError(Error):
    """Duplicate statement name specified within a script.

    Is only raised if a duplicate statement name exists **and** a method is
    invoked that tries to access statement(s) by their name as opposed to index
    position.

    """

    def __init__(
        self,
        nm: str,
        msg: Optional[str] = None,
        errno: Optional[int] = None,
        to_raise: Optional[bool] = False,
    ):
        super().__init__(nm=nm, msg=msg, errno=errno, to_raise=to_raise)

    def __str__(self):
        """DuplicateTagError message."""
        str_args = self.format_error_args(
            _filter=True,
            **{"name": f"'{self.nm}'", "msg": self.msg, "errno": self.errno},
        ).strip("\n")
        return f"""
indistinct statement names found within {self.nm}; wrap names must be unique if
running `script.contents(by_index=False)`.
see contents of `script.duplicates` for the exact wrap names causing the issue.
"""


class LoadingInternalError(InternalError):
    """Exception boundaries for :class:`~snowmobile.core.table.Table`."""

    def __init__(
        self,
        msg: Optional[str] = None,
        errno: Optional[int] = None,
        nm: Optional[str] = None,
        to_raise: Optional[bool] = False,
    ):
        super().__init__(nm=nm, msg=msg, errno=errno, to_raise=to_raise)

    def __str__(self):
        """LoadingInternalError message."""
        str_args = self.format_error_args(
            _filter=True, **{"name": self.nm, "msg": self.msg}
        ).strip("\n")
        return f"""
An internal exception was encountered in `snowmobile.Table`.
{str_args}
"""


class ExistingTableError(Error):
    """Table exists and `if_exists=Fail`"""
    pass


class ColumnMismatchError(Error):
    """Columns do not match and `if_exists!='replace'`"""
    pass


class FileFormatNameError(StatementNotFoundError):
    """The name of the pr file format is invalid."""
    pass


class InvalidTagsError(Error):
    """Error to be raised when a given statement wrap is not valid."""
    pass


class SnowflakeObjectNotFound(Error):
    """Error to be raised when a schema object is not found by snowmobile.core.sql."""
    pass


db_errors = (sfDatabaseError, ProgrammingError, pdDataBaseError)

snowmobile_errors = (
    Error,
    InternalError,
    StatementNotFoundError,
    StatementInternalError,
    StatementPostProcessingError,
    QAFailure,
    QADiffFailure,
    QAEmptyFailure,
    SnowFrameInternalError,
    StatementNotFoundError,
    DuplicateTagError,
    LoadingInternalError,
    ExistingTableError,
    ColumnMismatchError,
    FileFormatNameError,
    SnowflakeObjectNotFound,
)



"""
Derived :class:`~snowmobile.core.statement.Statement` classes.

These objects derive from :class:`snowmobile.core.statement.Statement`
and override its :meth:`~snowmobile.core.statement.Statement.process()`
method to perform additional post-processing of the statement's results
in conjunction with any parameters pr within the statement's
:ref:`tags <wrap>`.

``s.process()`` modifies a statement's
:attr:`~snowmobile.core.statement.Statement.outcome` attribute
(:class:`bool`) on which an assertion is run before continuing execution
of the script.

.. note::
    The ``on_exception`` and ``on_failure`` parameters of
    :meth:`script.run() <snowmobile.core.script.Script.run()>` are passed
    directly and only applicable to these derived statement classes.

        ``on_exception`` is used to control the exception-handling **of errors
        encountered in the post-processing invoked by**
        :meth:`s.process() <snowmobile.core.statement.Statement.process()>`

        ``on_failure`` is used to control the exception-handling **of a failed
        assertion ran on the outcome of the post-processing invoked by**
        :meth:`s.process() <snowmobile.core.statement.Statement.process()>`

"""
from __future__ import annotations

from typing import Any, Dict, List, Set

import pandas as pd

from .connection import Snowmobile  # isort: skip
from . import Statement  # isort: skip
from . import errors  # isort: skip


class QA(Statement):
    """Base class for QA st."""

    MSG = "object-specific exception message goes here."

    def __init__(self, sn: Snowmobile, **kwargs):
        super().__init__(sn=sn, **kwargs)

    def set_outcome(self):
        """Updates ._outcome upon completion of processing invoked by .process().

        """
        # fmt: off
        if self.e.outcome in [-1, 1]:
            return self

        # -- otherwise --
        if self.outcome:  # passed QA check
            self.e.set(outcome=-3)
            self._outcome = -3
        else:             # failed QA check
            self._outcome = -2
            self.e.set(outcome=-2)
            self.e.collect(
                e=self.DERIVED_FAILURE_MAPPING[self.anchor()](
                    nm=self.nm(),
                    msg=self.MSG,
                    idx=self.index,
                    to_raise=True,
                ),
            )
        # fmt: on

        return self


class Empty(QA):
    """QA class for verification that a statement's results are empty.

    The most widely applicable use of :class:`Empty` is for simple verification
    that a table's dimensions are as expected.

    """

    MSG = (
        "a 'qa-empty' check did not pass its validation; this means that "
        "a query you expected to return empty results returned a non-zero "
        "number of records."
    )

    def __init__(self, sn: Snowmobile, **kwargs):
        super().__init__(sn=sn, **kwargs)

    def process(self) -> QA:
        """Over-ride method; checks if results are empty and updates outcome"""
        self.outcome = self.results.empty
        return self.set_outcome()


class Diff(QA):
    """QA class for comparison of values within a table based on
    partitioning on a field.

    Attributes:
        partition_on (str):
            Column name to partition data on before comparing the
            partitioned datasets; defaults to 'src_description`.
        end_index_at (str):
            Column name that marks the last column to use as an index column
            when joining the partitioned datasets back together.
        compare_patterns (list):
            Regex patterns to match columns on that should be *included* in
            comparison (numeric columns you're running QA on).
        ignore_patterns (list):
            Regex patterns to match columns on that should be *ignored* both
            for the comparison and the index.
        generic_metric_col_nm (str):
            Column name to use for the melted field names; defaults to 'Metric'.
        compare_cols (list):
            Columns that are used in comparison once statement is executed
            and parsing is applied.
        drop_cols (list):
            Columns that are dropped once statement is executed and parsing
            is applied.
        idx_cols (list):
            Columns that are used for the index to join the data back
            together once statement is executed and parsing is applied.
        ub_raw (float):
            Maximum absolute raw difference (upper bound) that two fields
            that are being compared can differ from each other without
            causing a failure.
        ub_perc (float):
            Maximum absolute percentage difference (upper bound) that two
            comparison fields can differ from each other without causing a
            failure.

    """

    MSG = (
        "a 'qa-diff' check did not pass its validation; this means that "
        "the difference in the values compared did not fall within the "
        "specified threshold."
    )

    def __init__(self, sn: Snowmobile = None, **kwargs):
        """Instantiates a ``qa-diff`` statement.

        Args:
            delta_column_suffix (str):
                Suffix to add to columns that comparison is being run on;
                defaults to 'Delta'.
            partition_on (str):
                Column to partition the data on in order to compare.
            end_index_at (str):
                Column name that marks the last column to use as an index
                when joining the partitioned datasets back together.
            compare_patterns (list):
                Regex patterns matching columns to be *included* in
                comparison.
            ignore_patterns (list):
                Regex patterns to match columns on that should be *ignored*
                both for the comparison and the index.
            generic_metric_col_nm (str):
                Column name to use for the melted field names; defaults to
                'Metric'.
            raw_upper_bound (float):
                Maximum absolute raw difference that two fields that are
                being compared can differ from each other without causing a
                failure.
            percentage_upper_bound (float):
                Maximum absolute percentage difference that two comparison
                fields can differ from each other without causing a failure.

        """
        super().__init__(sn=sn, **kwargs)
        # fmt: off
        self._post_parse_init()
        self.compare_cols: List[str] = list()
        self.drop_cols: List[str] = list()
        self.idx_cols: List[str] = list()
        self.partitions: Dict[Any, pd.DataFrame] = dict()
        self.diff: bool = bool()
        # fmt: on

    def _post_parse_init(self) -> None:
        """Instantiates :class:`Diff` specific attributes.

        Combines arguments declared for the statement within the sql script
        and combines with defaults from ``snowmobile.toml``.

        """
        qa_cfg = self.sn.cfg.script.qa
        attrs_parsed = self.tag()
        self.partition_on: str = (
            attrs_parsed.get("partition-on", qa_cfg.partition_on)
        )
        self.end_index_at: str = (
            attrs_parsed.get("end-index-at", qa_cfg.end_index_at)
        )
        self.compare_patterns: List = (
            attrs_parsed.get("compare-patterns", qa_cfg.compare_patterns)
        )
        self.ignore_patterns: List = (
            attrs_parsed.get("ignore-patterns", qa_cfg.ignore_patterns)
        )
        self.relative_tolerance = attrs_parsed.get(
            "relative-tolerance", qa_cfg.tolerance.relative
        )
        self.absolute_tolerance = (
            attrs_parsed.get("absolute-tolerance", qa_cfg.tolerance.absolute)
            if not self.relative_tolerance
            else 0
        )

    def _idx(self) -> None:
        """Isolates columns to use as indices."""
        self.idx_cols = self.results.snf.cols_ending(
            self.end_index_at, ignore_patterns=[self.partition_on]
        )
        if not self.idx_cols:
            self.e.collect(
                e=errors.StatementPostProcessingError(
                    msg=(
                        f"Arguments pr don't result in any index columns "
                        f"on which to join DataFrame's partitions."
                    ),
                    to_raise=True,
                )
            ).set(outcome=-1)

    def _drop(self) -> None:
        """Isolates columns to ignore/drop."""
        self.drop_cols = self.results.snf.cols_matching(patterns=self.ignore_patterns)

    def _compare(self) -> None:
        """Isolate columns to use for comparison."""
        self.compare_cols = self.results.snf.cols_matching(
            patterns=self.compare_patterns,
            ignore_patterns=[self.partition_on, self.idx_cols, self.drop_cols],
        )
        if not self.compare_cols:
            self.e.collect(
                e=errors.StatementPostProcessingError(
                    msg=f"Arguments pr don't result in any comparison columns."
                ),
                to_raise=True,
            ).set(outcome=-1)

    def split_cols(self) -> Diff:
        """Post-processes results returned from a ``qa-diff`` statement.

        Executes private methods to split columns into:
            * Index columns
            * Drop columns
            * Comparison columns

        Then runs checks needed to ensure minimum requirements are met in order
        for a valid partition/comparison to be made.

        """
        try:
            self._idx()
            self._drop()
            self._compare()
        except Exception:
            raise

        # fmt: off
        if self.partition_on not in list(self.results.columns):
            self.e.collect(
                e=errors.StatementPostProcessingError(
                    msg=(
                        f"Column `{self.partition_on}` not found in DataFrames columns."
                    ),
                    to_raise=True,
                ),
            ).set(outcome=-1)
        # fmt: on

        return self

    @property
    def partitioned_by(self) -> Set[Any]:
        """Distinct values within the ``partition_on`` column that data is
        partitioned by.

        """
        return set(self.results[self.partition_on])

    @staticmethod
    def partitions_are_equal(
        partitions: Dict[str, pd.DataFrame], abs_tol: float, rel_tol: float
    ) -> bool:
        """Evaluates if a dictionary of DataFrames are identical.

        Args:
            partitions (Dict[str, pd.DataFrame]):
                A dictionary of DataFrames returned by
                :meth:`snowmobile.DataFrame`.
            abs_tol (float):
                Absolute tolerance for difference in any value amongst the
                DataFrames being compared.
            rel_tol (float):
                Relative tolerance for difference in any value amongst the
                DataFrames being compared.

        Returns (bool):
            Indication of equality amongst all the DataFrames contained in
            ``partitions``.

        """
        partitions_by_index: Dict[int, pd.DataFrame] = {
            i: df for i, df in enumerate(partitions.values(), start=1)
        }
        checks_for_equality: List[bool] = [
            partitions_by_index[i].snf.df_diff(
                df2=partitions_by_index[i + 1], rel_tol=rel_tol, abs_tol=abs_tol
            )
            for i in range(1, len(partitions_by_index))
        ]
        return all(checks_for_equality)

    def process(self) -> Diff:
        """Post-processing for :class:`Diff`-specific results.
        """
        try:
            self.outcome = False

            self.split_cols()
            if self.drop_cols:
                self.results.drop(columns=self.drop_cols, inplace=True)
            self.results.set_index(keys=self.idx_cols, inplace=True)

            try:
                self.partitions = self.results.snf.partitions(on=self.partition_on)
            except Exception as e:
                self.e.collect(
                    e=errors.StatementPostProcessingError(
                        msg=(e.args[0]), to_raise=True
                    )
                ).set(outcome=-1)

            if self.e.outcome != -1:
                # TESTS: add test to verify what happens if this fails
                self.outcome = self.partitions_are_equal(
                    partitions=self.partitions,
                    abs_tol=self.absolute_tolerance,
                    rel_tol=self.relative_tolerance,
                )

        finally:
            # TODO: Build in a protected message argument to statement tags
            #   so a custom message can be embedded per-check to display
            #   exactly what is being tested to stdout/logs when it fails.
            return self.set_outcome()



"""
Demonstrate specifying an alternate snowmobile.toml file *name*.
../docs/snippets/snowmobile/specifying_configuration2.py
"""
# -- SETUP --------------------------------------------------------------------

import time
import shutil
import snowmobile

# Instantiate sn from snowmobile.toml; omit unnecessary connection
sn = snowmobile.connect(delay=True)

# Create alternate snowmobile.toml file called 'snowmobile2.toml'
path_cfg_orig = sn.cfg.location
path_cfg2 = path_cfg_orig.parent / 'snowmobile2.toml'
shutil.copy(path_cfg_orig, path_cfg2)


# -- EXAMPLE ------------------------------------------------------------------

def alt_sn(n: int) -> snowmobile.Snowmobile:
    """Instantiate sn from snowmobile2.toml and print time elapsed."""
    pre = time.time()
    sn = snowmobile.connect(
        config_file_nm='snowmobile2.toml',
        delay=True  # omit connection - not needed
    )
    print(f"n={n}, time-required: ~{int(time.time() - pre)} seconds")
    return sn


sn_alt1 = alt_sn(n=1)  #> n=1, time-required: ~6 seconds  -> locates file, caches path
sn_alt2 = alt_sn(n=2)  #> n=2, time-required: ~0 seconds  -> uses cache from sn_alt1
"""
Note:
    The time required for `sn_alt1` to locate 'snowmobile2.toml' is arbitrary and
    will vary based the file's location relative to the current working directory.
"""


# -- TEARDOWN -----------------------------------------------------------------
# Deleting 'snowmobile2.toml' from file system post-example

import os
os.remove(sn_alt1.cfg.location)
# snowmobile-include



"""
Verify credentials order and alias is as laid out as defined in
snowmobile.Snowmobile intro assumptions.
..docs/snippets/connector_aliases.py
"""
import snowmobile

sn = snowmobile.Snowmobile()
print(sn.cfg.connection.current)  # > Credentials('creds1')

assert list(sn.cfg.connection.credentials)[:2] == [
    "creds1",
    "creds2",
], "Order of credentials is not as expected"


with open(sn.cfg.location, "r") as r:
    snowmobile_toml_raw = r.read()

assert (
    "default-creds = ''" in snowmobile_toml_raw
), "Value of 'default-creds' has been altered from default"

# -- complete example; should run 'as is' --
